SystemOrganization addCategory: #CodeImport!
SystemOrganization addCategory: #'CodeImport-Tests'!

!ClassCommentReader methodsFor: '*CodeImport' stamp: 'GuillermoPolito 5/5/2012 02:13'!
importCodeFrom: aCodeImporter
	aCodeImporter importClassComment: self.! !

!ClassCommentReader methodsFor: '*CodeImport' stamp: 'GuillermoPolito 5/5/2012 02:33'!
setClassComment: aClassComment
	self theClass theNonMetaClass
						classComment: aClassComment
						stamp: changeStamp! !

!ClassCategoryReader methodsFor: '*CodeImport' stamp: 'GuillermoPolito 5/5/2012 02:32'!
compileSourceCode: someSourceCode
	self theClass
			compile: someSourceCode 
			classified: category
			withStamp: changeStamp
			notifying: nil! !

!ClassCategoryReader methodsFor: '*CodeImport' stamp: 'GuillermoPolito 5/5/2012 02:13'!
importCodeFrom: aCodeImporter
	aCodeImporter importClassCategory: self.! !

!Categorizer methodsFor: '*CodeImport' stamp: 'GuillermoPolito 5/5/2012 02:12'!
importCodeFrom: aCodeImporter
	aCodeImporter importCategorizer: self.! !

Object subclass: #ChunkFileFormatParser
	instanceVariableNames: 'readStream parsedDeclarations nextChar'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CodeImport'!

!ChunkFileFormatParser commentStamp: '<historical>' prior: 0!
I represent the chunk file format.  Right now I know how to parse chunks from a readStream, but not how to write them.!

!ChunkFileFormatParser class methodsFor: 'instance creation' stamp: 'GuillermoPolito 5/5/2012 17:06'!
for: aReadStream
	^self new
		readStream: aReadStream;
		yourself! !

!ChunkFileFormatParser methodsFor: 'accessing' stamp: 'GuillermoPolito 5/5/2012 17:19'!
addDeclaration: aDeclaration
	parsedDeclarations	 add: aDeclaration ! !

!ChunkFileFormatParser methodsFor: 'testing' stamp: 'GuillermoPolito 5/5/2012 22:06'!
initialize
	parsedDeclarations := OrderedCollection new.! !

!ChunkFileFormatParser methodsFor: 'testing' stamp: 'GuillermoPolito 5/5/2012 22:06'!
isChunkEnd
	nextChar := readStream next.
	^ nextChar = nil or: [ (nextChar = self terminatorMark and: [ readStream peek ~= self terminatorMark ]) ]! !

!ChunkFileFormatParser methodsFor: 'testing' stamp: 'GuillermoPolito 5/5/2012 17:09'!
isNextChunkMetaData
	"If the next chunk starts with $!!, it will be an expression that after evaluation returns an object who knows how to read some metadata. 
	I.e. A chunk for comment reading should be something like:
	
	!!SomeClass commentStamp: '<historical>' prior: 0!!
	
	The object returned by the expression of that tag must understand #scanFrom:
	"
	
	^readStream peekFor: $!!! !

!ChunkFileFormatParser methodsFor: 'testing' stamp: 'GuillermoPolito 5/5/2012 17:09'!
isNextStyleChunk
	"Style chunks are between $] and $["
	^readStream peek == $]! !

!ChunkFileFormatParser methodsFor: 'testing' stamp: 'GuillermoPolito 5/5/2012 22:05'!
next
	"If the char we read previously is a terminator mark, it is scaping the next one, so we skip it"
	nextChar = self terminatorMark 
		ifTrue: [ readStream next ].
	^nextChar! !

!ChunkFileFormatParser methodsFor: 'testing' stamp: 'GuillermoPolito 5/5/2012 22:04'!
nextChunk
	| out next |
	out := (String new: 1000) writeStream.
	readStream skipSeparators.
	[ self isChunkEnd ] whileFalse: [
		out nextPut: self next.
	].
	^out contents! !

!ChunkFileFormatParser methodsFor: 'parsing' stamp: 'GuillermoPolito 5/5/2012 22:24'!
parseClassOrganization: classOrganizationPreamble
	"The comment preable is an array with the following structure:
	If instance side method:
	#(#CodeImportTestCaseTestClass #reorganize)
	
	if class side:
	
	#(#CodeImportTestCaseTestClass #class #reorganize)
	
	"
	| behaviorName isMeta stamp |
	behaviorName := classOrganizationPreamble first.
	isMeta := (classOrganizationPreamble second = #reorganize) not.
	self addDeclaration: (ClassOrganizationDeclaration
							contents: self nextChunk
							behaviorName: behaviorName
							isMeta: isMeta)! !

!ChunkFileFormatParser methodsFor: 'parsing' stamp: 'GuillermoPolito 5/5/2012 22:26'!
parseCommentDeclaration: commentPreamble
	"The comment preable is an array with the following structure:
	If instance side method:
	#(#CodeImportTestCaseTestClass #commentStamp: '<historical>' #prior: 0)
	
	if class side:
	
	#(#CodeImportTestCaseTestClass #class #commentStamp: '<historical>' #prior: 0)
	allButFirst:
	"
	| behaviorName isMeta stamp |
	behaviorName := commentPreamble first asSymbol.
	isMeta := (commentPreamble second = #commentStamp:) not.
	stamp := isMeta ifTrue: [commentPreamble at: 4] ifFalse: [commentPreamble at: 3].
	
	self addDeclaration: (ClassCommentDeclaration
							contents: self nextChunk
							behaviorName: behaviorName
							isMeta: isMeta
							stamp: stamp)! !

!ChunkFileFormatParser methodsFor: 'parsing' stamp: 'GuillermoPolito 5/5/2012 17:17'!
parseDeclarations
	[ readStream atEnd ] whileFalse: [
		self parseNextDeclaration.
	].
	^parsedDeclarations! !

!ChunkFileFormatParser methodsFor: 'parsing' stamp: 'GuillermoPolito 5/12/2012 14:05'!
parseMethodDeclarations: methodsPreamble
	| behaviorName isMeta category stamp methodSource |
	"The method preable is an array with the following structure:
	If instance side method:
	
	#(#CodeImportTestCaseTestClass #methodsFor: 'some protocol' #stamp: 'GuillermoPolito 5/2/2012 13:35')
	
	if class side:
	
	#(#CodeImportTestCaseTestClass #class #methodsFor: 'some protocol' #stamp: 'GuillermoPolito 5/2/2012 13:35')
	
	Sometimes there is no timestamp:
	
	#(#CodeImportTestCaseTestClass #methodsFor: 'some protocol')
	
	"
	
	behaviorName := methodsPreamble first.
	isMeta := (methodsPreamble second = #methodsFor:) not.
	category := isMeta 
					ifTrue: [ methodsPreamble at: 4 ]
					ifFalse: [ methodsPreamble at: 3 ].
	
	stamp := ''.
	methodsPreamble size > 4 ifTrue: [
		stamp := isMeta ifTrue: [ methodsPreamble at: 6 ] ifFalse: [ methodsPreamble at: 5 ].
	].
	
	[
		methodSource := self nextChunk.
		methodSource notEmpty
	] whileTrue: [ 
		self addDeclaration: (MethodDeclaration
								contents: methodSource
								behaviorName: behaviorName asSymbol
								isMeta: isMeta
								category: category
								stamp: stamp).
	]! !

!ChunkFileFormatParser methodsFor: 'parsing' stamp: 'GuillermoPolito 5/5/2012 22:28'!
parseNextDeclaration
	| isMetadata nextChunk value |
	readStream skipSeparators.
	
	(self isNextStyleChunk) ifTrue: [
		self addDeclaration: (StyleDeclaration contents: (self readUpToEndOfStyleChunk)).
		^self
	].

	isMetadata := self isNextChunkMetaData.
	nextChunk := self nextChunk.
	
	isMetadata
		ifFalse: [ 
			self addDeclaration: (DoItDeclaration contents: nextChunk).
		] ifTrue: [
			| substrings |
			substrings := Scanner new scanTokens: nextChunk.
			
			(substrings includes: 'methodsFor:')
				ifTrue: [ ^self parseMethodDeclarations: substrings ].
			(substrings includes: 'commentStamp:')
				ifTrue: [ ^self parseCommentDeclaration: substrings ].
			(substrings includes: 'reorganize')
				ifTrue: [ ^self parseClassOrganization: substrings ].
		]! !

!ChunkFileFormatParser methodsFor: 'accessing' stamp: 'GuillermoPolito 5/5/2012 17:07'!
readStream: aReadStream
	readStream := aReadStream! !

!ChunkFileFormatParser methodsFor: 'parsing' stamp: 'GuillermoPolito 5/5/2012 17:08'!
readUpToEndOfStyleChunk
	"Style chunks are between $] and $["
	^readStream upTo: $[! !

!ChunkFileFormatParser methodsFor: 'parsing' stamp: 'GuillermoPolito 5/5/2012 21:32'!
terminatorMark
	^$!!! !

Object subclass: #CodeDeclaration
	instanceVariableNames: 'contents'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CodeImport'!

!CodeDeclaration commentStamp: '<historical>' prior: 0!
I'm the abstract superclass of all the code declarations that should exist in an exported file.!

CodeDeclaration subclass: #BehaviorInstallingDeclaration
	instanceVariableNames: 'behaviorName isMeta'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CodeImport'!

!BehaviorInstallingDeclaration commentStamp: '<historical>' prior: 0!
I'm an abstract class grouping the behavior of the code declarations that will be instaled in a behavior, such as methods, class comments, organizations...!

!BehaviorInstallingDeclaration methodsFor: 'accessing' stamp: 'GuillermoPolito 5/5/2012 20:35'!
behaviorName: aBehaviorName
	behaviorName := aBehaviorName! !

!BehaviorInstallingDeclaration methodsFor: 'testing' stamp: 'GuillermoPolito 5/5/2012 20:37'!
existsBehavior
	^self class environment includesKey: behaviorName! !

!BehaviorInstallingDeclaration methodsFor: 'accessing' stamp: 'GuillermoPolito 5/5/2012 20:35'!
isMeta: aBoolean
	isMeta := aBoolean! !

!BehaviorInstallingDeclaration methodsFor: 'accessing' stamp: 'GuillermoPolito 5/5/2012 20:36'!
targetClass
	| target |
	target := self class environment at: behaviorName.
	^isMeta
		ifTrue: [ target classSide ]
		ifFalse: [ target ]! !

BehaviorInstallingDeclaration subclass: #ClassCommentDeclaration
	instanceVariableNames: 'stamp'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CodeImport'!

!ClassCommentDeclaration commentStamp: '<historical>' prior: 0!
I represent the declaration of a class comment in a file.  My contents are the class comment to import, and the comment reader is the object who knows the class where to put the class comment.

Sending me the message #import makes me install the comment into the class.!

!ClassCommentDeclaration class methodsFor: 'instance creation' stamp: 'GuillermoPolito 5/5/2012 21:04'!
contents: someContents behaviorName: behaviorName isMeta: aBoolean stamp: stamp
	^self new
			contents: someContents;
			behaviorName: behaviorName;
			isMeta: aBoolean;
			stamp: stamp;
			yourself! !

!ClassCommentDeclaration methodsFor: 'importing' stamp: 'GuillermoPolito 5/5/2012 20:41'!
import
	(self existsBehavior)
		ifFalse: [ self error: ('Cannot install comment in unexistent behavior {1}' format: { behaviorName asString } ) ].
	
	^self targetClass theNonMetaClass
						classComment: contents
						stamp: stamp! !

!ClassCommentDeclaration methodsFor: 'accessing' stamp: 'GuillermoPolito 5/5/2012 20:37'!
stamp: classCommentStamp
	stamp := classCommentStamp! !

BehaviorInstallingDeclaration subclass: #ClassOrganizationDeclaration
	instanceVariableNames: 'classOrganizer'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CodeImport'!

!ClassOrganizationDeclaration commentStamp: '<historical>' prior: 0!
I represent the declaration of a class organization in a file.  My contents is the class organization string to import, and the class organizer is the one of the changed class.

Sending me the message #import makes me install the class organization I carry into the class.!

!ClassOrganizationDeclaration class methodsFor: 'instance creation' stamp: 'GuillermoPolito 5/5/2012 20:48'!
contents: someContents behaviorName: behaviorName isMeta: aBoolean
	^self new
		contents: someContents;
		behaviorName: behaviorName;
		isMeta: aBoolean;
		yourself! !

!ClassOrganizationDeclaration class methodsFor: 'instance creation' stamp: 'GuillermoPolito 5/5/2012 16:31'!
contents: someContents organizer: aClassOrganizer 
	^self new
			contents: someContents;
			organizer: aClassOrganizer;
			yourself! !

!ClassOrganizationDeclaration methodsFor: 'importing' stamp: 'GuillermoPolito 5/5/2012 20:44'!
import
	(self existsBehavior)
		ifFalse: [ self error: ('Cannot change organization of unexistent behavior {1}' format: { behaviorName asString } ) ]. 
		
	self targetClass organization changeFromString: contents! !

BehaviorInstallingDeclaration subclass: #MethodDeclaration
	instanceVariableNames: 'categoryReader categoryName stamp'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CodeImport'!

!MethodDeclaration commentStamp: '<historical>' prior: 0!
I represent the declaration of a method.  My contents are the source code to import, and the category reader is the object who knows the class, category and timestamp where to install the method.

Sending me the message #import makes me install the method into the class.!

!MethodDeclaration class methodsFor: 'instance creation' stamp: 'GuillermoPolito 5/5/2012 20:40'!
contents: someContents behaviorName: behaviorName isMeta: isMeta category: categoryName stamp: stamp
	^self new
			contents: someContents;
			behaviorName: behaviorName;
			isMeta: isMeta;
			category: categoryName;
			stamp: stamp;
			yourself! !

!MethodDeclaration methodsFor: 'accessing' stamp: 'GuillermoPolito 5/5/2012 20:13'!
category: aCategoryName
	categoryName := aCategoryName! !

!MethodDeclaration methodsFor: 'accessing' stamp: 'GuillermoPolito 5/5/2012 16:24'!
categoryReader: aClassCategoryReader 
	categoryReader := aClassCategoryReader! !

!MethodDeclaration methodsFor: 'importing' stamp: 'GuillermoPolito 5/5/2012 20:41'!
import
	(self existsBehavior)
		ifFalse: [ self error: ('Cannot install method in unexistent behavior {1}' format: { behaviorName asString } ) ].
	
	^self targetClass
		compile: contents
		classified: categoryName
		withStamp: stamp
		notifying: nil! !

!MethodDeclaration methodsFor: 'accessing' stamp: 'GuillermoPolito 5/5/2012 20:13'!
stamp: aMethodStamp
	stamp := aMethodStamp! !

!CodeDeclaration class methodsFor: 'instance creation' stamp: 'GuillermoPolito 5/5/2012 16:02'!
contents: someContents
	^self new
			contents: someContents;
			yourself! !

!CodeDeclaration methodsFor: 'accessing' stamp: 'GuillermoPolito 5/5/2012 16:02'!
contents
	^ contents! !

!CodeDeclaration methodsFor: 'accessing' stamp: 'GuillermoPolito 5/5/2012 16:02'!
contents: anObject
	contents := anObject! !

!CodeDeclaration methodsFor: 'printing' stamp: 'GuillermoPolito 5/5/2012 20:21'!
printOn: aStream
	super printOn: aStream.
	aStream nextPut: $(.
	aStream nextPutAll: contents.
	aStream nextPut: $).! !

CodeDeclaration subclass: #DoItDeclaration
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CodeImport'!

!DoItDeclaration commentStamp: '<historical>' prior: 0!
I represent the declaration of some code to evaluate.  My contents are the code to evaluate.

Sending me the message #import makes me evaluate the code and return the result of the computation.

Be careful, my #import method also handles ChangeSet interaction so far, handling Changeset preambles and Postscripts.!

!DoItDeclaration methodsFor: 'importing' stamp: 'GuillermoPolito 5/5/2012 17:14'!
import
	self flag: #fixme. "Ugly hack to parse preamble and postcript as a do it."
	((contents beginsWith: '"Change Set:') and: [ChangeSet current preambleString == nil])
		ifTrue: [ChangeSet current preambleString: contents].
	((contents beginsWith: '"Postscript:') and: [ChangeSet current postscriptString == nil])
		ifTrue: [ChangeSet current postscriptString: contents].

	^Compiler evaluate: contents logged: false.! !

CodeDeclaration subclass: #StyleDeclaration
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CodeImport'!

!StyleDeclaration commentStamp: '<historical>' prior: 0!
I'm the declaration of a style stuff into a file.  I'm not really used, but I'm here to make parsing an homogeneous process and backwards compatibility.

Sending me the #import message does nothing :).!

!StyleDeclaration methodsFor: 'importing' stamp: 'GuillermoPolito 5/5/2012 16:40'!
import
	"Nothing to do with style declarations"
	^''! !

Object subclass: #CodeImporter
	instanceVariableNames: 'readStream codeDeclarations parserClass'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CodeImport'!

!CodeImporter commentStamp: '<historical>' prior: 0!
I'm an object in charge of import source files.  I know a format object that knows how to parse the files, and I import the parsed results into the image.

I handle doIts normally, but some special cases like class organization, class comment or methods are handled via a double dispatch (See my method extensions for that).

=-=-=- How to use me -=-=-=

If you want to fileIn the code -aka compile/evaluate it:

CodeImporter evaluateReadStream: '2+2!!' readStream.

or

CodeImporter evaluateString: '2+2!!'

or

CodeImporter evaluateFileNamed: 'something.st'

or

CodeImporter evaluateFileStream: (FileStream readOnlyFileNamed: 'something.st')

Now, you can also generate a model of code declarations inside a file by just creating an instance:

CodeImporter fileStream: (FileStream readOnlyFileNamed: 'something.st').

And then query me sending the message

#codeDeclarations

Instances of me can be created through 
#fromString:
#readStream:
#fileNamed:
#fileStream:!

!CodeImporter class methodsFor: 'evaluating' stamp: 'GuillermoPolito 5/5/2012 17:23'!
evaluateFileNamed: aFileName
	^(self fileNamed: aFileName) evaluateDeclarations! !

!CodeImporter class methodsFor: 'evaluating' stamp: 'GuillermoPolito 5/5/2012 17:23'!
evaluateFileStream: aFileStream
	^(self fileStream: aFileStream) evaluateDeclarations! !

!CodeImporter class methodsFor: 'evaluating' stamp: 'GuillermoPolito 5/5/2012 17:25'!
evaluateReadStream: aReadStream
	^(self readStream: aReadStream) evaluateDeclarations! !

!CodeImporter class methodsFor: 'evaluating' stamp: 'GuillermoPolito 5/5/2012 17:25'!
evaluateString: aString
	^(self fromString: aString) evaluateDeclarations! !

!CodeImporter class methodsFor: 'instance creation' stamp: 'GuillermoPolito 5/5/2012 17:23'!
fileNamed: aFileName
	^self file: (FileStream readOnlyFileNamed: aFileName).
! !

!CodeImporter class methodsFor: 'instance creation' stamp: 'GuillermoPolito 5/5/2012 17:23'!
fileStream: aFileStream
	^self new
		file: aFileStream;
		yourself! !

!CodeImporter class methodsFor: 'instance creation' stamp: 'GuillermoPolito 5/5/2012 17:25'!
fromString: aByteString 
	^self readStream: aByteString readStream! !

!CodeImporter class methodsFor: 'instance creation' stamp: 'GuillermoPolito 5/5/2012 17:24'!
readStream: aReadStream
	^self new
		readStream: aReadStream;
		yourself! !

!CodeImporter methodsFor: 'accessing' stamp: 'GuillermoPolito 5/5/2012 16:12'!
codeDeclarations
	^codeDeclarations! !

!CodeImporter methodsFor: 'evaluating' stamp: 'GuillermoPolito 5/5/2012 16:45'!
evaluateDeclarations
	"Evaluates the declarations from the text in the file and answers the last result"

	| value |
	self parseDeclarations.
	self codeDeclarations do: [ :decl | value := decl import ].
	self flushChangesFile.
	^value! !

!CodeImporter methodsFor: 'accessing' stamp: 'GuillermoPolito 5/5/2012 01:39'!
file: aFileStream
	self readStream: aFileStream.
	self selectTextConverterForCode.! !

!CodeImporter methodsFor: 'evaluating' stamp: 'GuillermoPolito 5/5/2012 01:12'!
flushChangesFile
	"Yes, we need to do this in order to flush the file, puaj"
	Smalltalk logChange: '----End fileIn----'.! !

!CodeImporter methodsFor: 'initialize-release' stamp: 'GuillermoPolito 5/5/2012 17:06'!
initialize
	codeDeclarations := OrderedCollection new.
	parserClass := ChunkFileFormatParser.! !

!CodeImporter methodsFor: 'evaluating' stamp: 'GuillermoPolito 5/5/2012 17:17'!
parseDeclarations
	codeDeclarations := (parserClass for: readStream) parseDeclarations.! !

!CodeImporter methodsFor: 'accessing' stamp: 'GuillermoPolito 5/5/2012 01:39'!
readStream: aReadStream
	readStream := aReadStream.! !

!CodeImporter methodsFor: 'fileHandling' stamp: 'GuillermoPolito 5/5/2012 01:38'!
selectTextConverterForCode
	self flag: #fix.  "This should not be here probably."
	
	"We need to see the first three bytes in order to see the origin of the file"
	readStream binary.
	((readStream next: 3) = #[ 16rEF 16rBB 16rBF ]) ifTrue: [
		readStream converter: UTF8TextConverter new
	] ifFalse: [
		readStream converter: MacRomanTextConverter new.
	].

	"we restore the position to the start of the file again"
	readStream position: 0.
	
	"We put the file in text mode for the file in"
	readStream text.! !
